function obj_set_hitbox_custom(o, hitbox)
    if o == nil or hitbox == nil then return end

    if (o.oFlags & OBJ_FLAG_30) == 0 then
        o.oFlags = o.oFlags | OBJ_FLAG_30

        o.oInteractType = hitbox.interactType
        o.oDamageOrCoinValue = hitbox.damageOrCoinValue
        o.oHealth = hitbox.health
        o.oNumLootCoins = hitbox.numLootCoins

        cur_obj_become_tangible()
    end

    o.hitboxRadius = o.header.gfx.scale.x * hitbox.radius
    o.hitboxHeight = o.header.gfx.scale.y * hitbox.height
    o.hurtboxRadius = o.header.gfx.scale.x * hitbox.hurtboxRadius
    o.hurtboxHeight = o.header.gfx.scale.y * hitbox.hurtboxHeight
    o.hitboxDownOffset = o.header.gfx.scale.y * hitbox.downOffset
end

-------------------
-- BOWLING BALLS --
-------------------

function bowling_ball_set_hitbox(o)
    local sBowlingBallHitbox = {
        ["damageOrCoinValue"] = 2,
        ["downOffset"] = 0,
        ["health"] = o.oHealth,
        ["height"] = 150,
        ["hurtboxHeight"] = 0,
        ["hurtboxRadius"] = 0,
        ["interactType"] = INTERACT_DAMAGE,
        ["numLootCoins"] = 0,
        ["radius"] = 100,
    }

    obj_set_hitbox_custom(o, sBowlingBallHitbox)

    if (o.oInteractStatus & INT_STATUS_INTERACTED) ~= 0 then
        o.oInteractStatus = 0
    end

    if o.oHealth <= 0 then
        bowling_ball_destroy(o)
    end

    if o.oForwardVel > 64 then
        o.oForwardVel = 64
    end
end

function bowling_ball_destroy(o)
    spawn_mist_particles()
    spawn_mist_particles_variable(0, 0, 92)

    cur_obj_play_sound_2(SOUND_GENERAL_QUIET_POUND1_LOWPRIO)
    set_camera_shake_from_point(SHAKE_POS_BOWLING_BALL, o.oPosX, o.oPosY, o.oPosZ)
    obj_mark_for_deletion(o)
end

id_bhvKingBowlingBall = hook_behavior(nil, OBJ_LIST_GENACTOR, true, function(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    o.header.gfx.node.flags = o.header.gfx.node.flags | GRAPH_RENDER_BILLBOARD
    o.oGraphYOffset = 130

    o.oGravity = 12
    o.oFriction = 1
    o.oBuoyancy = 2
    o.oHealth = 16
end,

function(o)
    local collisionFlags = object_step()
    bowling_ball_set_hitbox(o)

    if (collisionFlags & OBJ_COL_FLAG_HIT_WALL) ~= 0 then
        cur_obj_play_sound_2(SOUND_OBJ_BULLY_METAL)
        set_camera_shake_from_point(SHAKE_POS_BOWLING_BALL, o.oPosX, o.oPosY, o.oPosZ)

        if (collisionFlags & OBJ_COL_FLAG_GROUNDED) ~= 0 then
            o.oVelY = 64
        end

        o.oHealth = o.oHealth - 1
    end

    if (collisionFlags & OBJ_COL_FLAG_GROUNDED) ~= 0 then
        cur_obj_play_sound_1(SOUND_ENV_UNKNOWN2)

        if o.oVelY > 5 then
            cur_obj_play_sound_2(SOUND_GENERAL_QUIET_POUND1_LOWPRIO)
            set_camera_shake_from_point(SHAKE_POS_BOWLING_BALL, o.oPosX, o.oPosY, o.oPosZ)
        end
    end

    set_object_visibility(o, 3000)
end)

if gKingStates == nil then return end
if gKingStates and gKingStates[0].isRomHack then return end

------------------
-- GRAB HANDLER --
------------------

hook_event(HOOK_MARIO_UPDATE, function(m)
    local gSync = gPlayerSyncTable[m.playerIndex]
    local k = gKingStates[m.playerIndex]

    if m.playerIndex == 0 and gSync.isKing then
        if m.heldObj ~= nil and (obj_has_behavior_id(m.heldObj, id_bhvMips) ~= 0 or (obj_has_behavior_id(m.heldObj, id_bhvUkiki) ~= 0 and m.heldObj.oBehParams2ndByte == 1)) then
            m.actionTimer = 0
            m.marioBodyState.grabPos = GRAB_POS_LIGHT_OBJ

            if m.action == ACT_PLACING_DOWN then
                m.heldObj.oKoopaMovementType = 1              -- i use koopa movement type for the special "kill" code as i highly doubt any
                set_mario_action(m, ACT_PICKING_UP_BOWSER, 0)       -- other mods or hacks will apply such a strange value to mips or ukiki
            end
        end

        if m.action == ACT_HOLDING_BOWSER then
            k.bowserAngleVel = math.abs(m.angleVel.y) / 100
        elseif m.action ~= ACT_RELEASING_BOWSER then
            k.bowserAngleVel = 0
        end
    end
end)

----------
-- MIPS --
----------

hook_behavior(id_bhvMips, OBJ_LIST_GENACTOR, false, function(o)
    o.oKoopaMovementType = 0
end,

function(o)
    local i = network_local_index_from_global(o.globalPlayerIndex)
    local gSync = gPlayerSyncTable[i]
    local k = gKingStates[i]

    if gSync.isKing and o.oKoopaMovementType == 1 then
        if o.oAction == MIPS_ACT_FALL_DOWN then
            if o.oTimer == 0 then
                o.oGravity = 2.5
                o.oForwardVel = max(k.bowserAngleVel, 5)
                o.oVelY = max(k.bowserAngleVel, 5)
            end

            o.oTimer = o.oTimer + 1
        end

        if o.oAction == MIPS_ACT_WAIT_FOR_ANIMATION_DONE then
            spawn_mist_particles()
            spawn_triangle_break_particles(20, 138, 0.7, 3)
            obj_spawn_yellow_coins(o, 3)
            play_sound(SOUND_GENERAL_BREAK_BOX, o.header.gfx.cameraToObject)
            cur_obj_disable_rendering_and_become_intangible(o)

            o.oAction = MIPS_ACT_IDLE
        end
    end
end)

-----------
-- UKIKI --
-----------

hook_behavior(id_bhvUkiki, OBJ_LIST_GENACTOR, false, function(o)
    o.oKoopaMovementType = 0
end,

function(o)
    local i = network_local_index_from_global(o.globalPlayerIndex)
    local gSync = gPlayerSyncTable[i]
    local k = gKingStates[i]
    local m = gMarioStates[i]

    if gSync.isKing then
        if o.oBehParams2ndByte == 1 and o.oKoopaMovementType == 1 then
            --if o.oPosY ~= o.oFloorHeight and o.oAction ~= UKIKI_ACT_JUMP then
            if m.heldObj ~= o then
                cur_obj_enable_rendering()
                o.header.gfx.node.flags = o.header.gfx.node.flags & ~GRAPH_RENDER_INVISIBLE
                o.oHeldState = HELD_FREE
                o.oFlags = o.oFlags & ~OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW
                cur_obj_init_animation(UKIKI_ANIM_JUMP)
                cur_obj_become_tangible()
                o.oForwardVel = max(k.bowserAngleVel, 5)
                o.oVelY = max(k.bowserAngleVel, 5)
                o.oAction = UKIKI_ACT_JUMP
                o.oKoopaMovementType = 2
            end
        end
    else
        o.oKoopaMovementType = 0
    end

    if o.oAction == UKIKI_ACT_JUMP and o.oKoopaMovementType == 2 then
        if o.oPosY == o.oFloorHeight then
            spawn_mist_particles()
            spawn_triangle_break_particles(20, 138, 0.7, 3)
            obj_spawn_yellow_coins(o, 3)
            play_sound(SOUND_GENERAL_BREAK_BOX, o.header.gfx.cameraToObject)
            --cur_obj_disable_rendering_and_become_intangible(o)

            if o.oAnimState == 2 then
                local cap = spawn_non_sync_object(id_bhvNormalCap, E_MODEL_MARIOS_CAP, o.oPosX, o.oPosY, o.oPosZ, nil)
                if cap ~= nil then
                    cap.oVelY = 20
                end
            end

            obj_mark_for_deletion(o)
        end
    end
end)

---------------------
-- KOOPA THE QUICK --
---------------------

hook_behavior(id_bhvKoopa, OBJ_LIST_PUSHABLE, false, nil, function(o)
    local i = network_local_index_from_global(o.globalPlayerIndex)
    local gSync = gPlayerSyncTable[i]
    local k = gKingStates[i]
    local m = gMarioStates[i]

    if gSync.isKing then
        if o.parentObj then
            if k.cheatKoopa == 0
            and ((m.flags & MARIO_WING_CAP) ~= 0 or (m.flags & MARIO_TELEPORTING) ~= 0 or (hud_get_value(HUD_DISPLAY_TIMER)/30 <= 3 and o.parentObj.oKoopaRaceEndpointRaceStatus == 1)) then
                k.cheatKoopa = 1
            end

            if k.cheatKoopa == 1 then
                k.cheatKoopa = 1
                o.parentObj.oKoopaRaceEndpointRaceStatus = -1
            end

            if o.parentObj.oKoopaRaceEndpointUnk100 == DIALOG_006 then
                k.cheatKoopa = 2
                o.parentObj.oKoopaRaceEndpointRaceStatus = 0
            end
        end

        if o.oKoopaMovementType >= KOOPA_BP_KOOPA_THE_QUICK_BASE then
            if o.oAction == KOOPA_THE_QUICK_ACT_RACE and o.oSubAction == KOOPA_THE_QUICK_SUB_ACT_RUN then
                local downhillSteepness = 1 + sins(o.oPathedTargetPitch)

                o.oKoopaAgility = 10
                obj_forward_vel_approach(o.oKoopaAgility * 6.0 * downhillSteepness, o.oKoopaAgility * 0.1)
            end
        end
    end
end)

----------------
-- INTRO KING --
----------------

hook_behavior(id_bhvBeginningPeach, OBJ_LIST_DEFAULT, false, nil, function(o)
    smlua_anim_util_set_animation(o, "KING_ANIM_STAND")

    if obj_has_model_extended(o, E_MODEL_KING_ALONE) == 0 then
        obj_set_model_extended(o, E_MODEL_KING_ALONE)
    end

    o.oPosZ = o.oPosZ - 32
    o.oPosX = o.oPosX - 122
    cur_obj_scale(0.3)
end)