if gKingStates == nil then return end
if gKingStates and gKingStates[0].isRomHack then return end

KING_NPC_ACT_IDLE = 0
KING_NPC_ACT_TURN_TO_TALK = 1
KING_NPC_ACT_TALK = 2

function bhv_king_npc_init(o)
    o.oFlags = OBJ_FLAG_COMPUTE_ANGLE_TO_MARIO | OBJ_FLAG_COMPUTE_DIST_TO_MARIO | OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW | OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    o.oBobombBuddyRole = 0

    o.oGravity = 2.5
    o.oFriction = 0.8
    o.oBuoyancy = 1.3

    o.hitboxRadius = 100
    o.hitboxHeight = 60

    o.oInteractionSubtype = INT_SUBTYPE_NPC

    o.oMoveAngleYaw = 0x8000

    obj_init_animation(o, MARIO_ANIM_FIRST_PERSON)
    smlua_anim_util_set_animation(o, "KING_ANIM_STAND")
end

function king_npc_act_idle(o)
    o.oBobombBuddyPosXCopy = o.oPosX
    o.oBobombBuddyPosYCopy = o.oPosY
    o.oBobombBuddyPosZCopy = o.oPosZ

    object_step()

    if o.oDistanceToMario < 1000 then
        o.oMoveAngleYaw = approach_s16_symmetric(o.oMoveAngleYaw, o.oAngleToMario, 0x140)
    else
        o.oMoveAngleYaw = approach_s16_symmetric(o.oMoveAngleYaw, 0x8000, 0x140)
    end

    if o.oInteractStatus == INT_STATUS_INTERACTED then
        o.oAction = KING_NPC_ACT_TURN_TO_TALK
    end
end

function king_npc_act_talk(o)
    create_dialog_box_with_response(DIALOG_099)
    o.activeFlags = o.activeFlags | ACTIVE_FLAG_INITIATED_TIME_STOP

    if get_dialog_box_state() == DIALOG_UNK2_END_DIALOG then
        o.activeFlags = o.activeFlags & ~ACTIVE_FLAG_INITIATED_TIME_STOP
        o.oBobombBuddyHasTalkedToMario = BOBOMB_BUDDY_HAS_TALKED
        o.oInteractStatus = 0
        o.oAction = KING_NPC_ACT_IDLE
    end
end

function king_npc_act_turn_to_talk(o)
    o.oMoveAngleYaw = approach_s16_symmetric(o.oMoveAngleYaw, o.oAngleToMario, 0x1000)
    if convert_s16(o.oMoveAngleYaw) == convert_s16(o.oAngleToMario) then
        o.oAction = KING_NPC_ACT_TALK
    end
end

function bhv_king_npc_loop(o)
    if o.oAction == KING_NPC_ACT_IDLE then
        king_npc_act_idle(o)
    elseif o.oAction == KING_NPC_ACT_TURN_TO_TALK then
        king_npc_act_turn_to_talk(o)
    elseif o.oAction == KING_NPC_ACT_TALK then
        king_npc_act_talk(o)
    end

    set_object_visibility(o, 3000)
    cur_obj_push_mario_away_from_cylinder(o.hitboxRadius, o.hitboxHeight)
    o.oInteractStatus = 0
end

id_bhvKingNPC = hook_behavior(nil, OBJ_LIST_GENACTOR, true, bhv_king_npc_init, bhv_king_npc_loop)

gKingDialogTimer = {}
for i = 0, (MAX_PLAYERS - 1) do
    gKingDialogTimer[i] = 0
end

function do_king_save_flag(m)
    if (save_file_get_flags() & SAVE_FLAG_TALKED_TO_KING) == 0 then

        play_puzzle_jingle()
        save_file_set_flags(SAVE_FLAG_TALKED_TO_KING)

        if m.specialTripleJump ~= 0 then
            m.specialTripleJump = 0
            djui_popup_create("|VV|       Yoshis reward       |VV|\n|__|   has now been removed   |__|", 2)
        end

        djui_popup_create_global("|VV|          You have         |VV|\n|__|       infinite lives       |__|", 2)
        djui_popup_create_global("|VV|    Now Stars will give    |VV|\n|__|      you invincibility      |__|", 2)

    else
        m.numLives = 4
        save_file_clear_flags(SAVE_FLAG_TALKED_TO_KING)
    end

    save_file_do_save(get_current_save_file_num() - 1, 1)
end

function mario_get_close_object(m, behaviorId)
    local o = obj_get_nearest_object_with_behavior_id(m.marioObj, behaviorId)

    if o then
        if dist_between_objects(m.marioObj, o) < 640
        and (o.activeFlags & ACTIVE_FLAG_ACTIVE) ~= 0 then
            return true
        end
    end

    return false
end

function mario_update(m)
    local gSync = gPlayerSyncTable[m.playerIndex]
    local k = gKingStates[m.playerIndex]

    if m.playerIndex == 0 then
        if not gSync.isKing then
            k.starKing.enabled = 0
        elseif k.starKing.timer ~= 0 then
            k.starKing.enabled = 1
        end

        k.starKing.timer = min(k.starKing.timer, 1500)
        k.starKing.timer = max(k.starKing.timer - 1, 0)

        if k.starKing.timer <= 0 then
            k.starKing.enabled = 0
        end

        if k.starKing.enabled == 1 then
            if ((get_global_timer() + 5) % 10) == 0 then
                spawn_sync_object(id_bhvSparkle, E_MODEL_SPARKLES_ANIMATION, m.pos.x, m.pos.y, m.pos.z, function(o)
                    obj_translate_xyz_random(o, 90)
                end)
            end

            m.marioBodyState.modelState = 0x400
        end

        if (save_file_get_flags() & SAVE_FLAG_TALKED_TO_KING) ~= 0 then
            if gSync.isKing then
                k.starKing.closeToStar = mario_get_close_object(m, id_bhvStar)
                or mario_get_close_object(m, id_bhvSpawnedStar)
                or mario_get_close_object(m, id_bhvSpawnedStarNoLevelExit)
                or mario_get_close_object(m, id_bhvUkikiCageStar)
                or mario_get_close_object(m, id_bhvStarSpawnCoordinates)

                if k.starKing.closeToStar and k.starKing.timer <= 1200 then
                    if ((m.input & INPUT_Z_PRESSED) ~= 0 and (m.controller.buttonDown & L_TRIG) ~= 0)
                    or ((m.input & INPUT_Z_DOWN) ~= 0 and (m.controller.buttonPressed & L_TRIG) ~= 0) then
                        play_sound(SOUND_MENU_STAR_SOUND, m.marioObj.header.gfx.cameraToObject)
                        k.starKing.timer = 1500
                        k.starKing.enabled = 1
                    end
                end
            end

            if m.numStars < 120 then
                m.numLives = 4
                save_file_clear_flags(SAVE_FLAG_TALKED_TO_KING)
            end
        end

        if get_dialog_id() == DIALOG_099
        and get_dialog_box_state() == DIALOG_UNK2_END_DIALOG
        and get_dialog_response() == 1 then

            if gKingDialogTimer[0] == 0 then
                do_king_save_flag(m)
            end

            gKingDialogTimer[0] = gKingDialogTimer[0] + 1
        else
            gKingDialogTimer[0] = 0
        end

        if m.numStars >= 120 and k.isRomHack == false then
            if gNetworkPlayers[0].currLevelNum == 6 then
                obj_mark_for_deletion(obj_get_first_with_behavior_id(id_bhvBooInCastle))
            end

            if gNetworkPlayers[0].currLevelNum == 26 then
                local bigBoo = obj_get_first_with_behavior_id(id_bhvBooWithCage)

                if bigBoo ~= nil then
                    spawn_sync_object(id_bhvBooCage, E_MODEL_HAUNTED_CAGE, bigBoo.oPosX, bigBoo.oPosY, bigBoo.oPosZ, nil)
                    obj_mark_for_deletion(bigBoo)
                end

                obj_mark_for_deletion(obj_get_first_with_behavior_id(id_bhvGhostHuntBoo))

                if obj_get_first_with_behavior_id(id_bhvKingNPC) == nil then
                    spawn_sync_object(id_bhvKingNPC, E_MODEL_KING_ALONE, 2716.7, 0, 561.1, nil)
                end
            end
        end
    end
end

function custom_mario_can_talk(m, o)
    if (m.action & ACT_FLAG_IDLE) ~= 0 or m.action == ACT_WALKING then
        local facingDYaw = mario_obj_angle_to_object(m, o) - m.faceAngle.y

        if facingDYaw >= -0x4000 and facingDYaw <= 0x4000 then
            return 1
        end
    end
             -- this should return 0 but idk how to cancel
    return 0   -- other actions before triggering dialogue
end

function custom_check_npc_talk(m, o)
    if custom_mario_can_talk(m, o) ~= 0 then

        if (m.controller.buttonPressed & A_BUTTON) ~= 0
        or (m.controller.buttonPressed & B_BUTTON) ~= 0 then

            o.oInteractStatus = INT_STATUS_INTERACTED

            m.interactObj = o
            m.usedObj = o

            m.pos.y = m.floorHeight
            return set_mario_action(m, ACT_READING_NPC_DIALOG, 0)
        end

    end
end

function before_mario_update(m)
    local gSync = gPlayerSyncTable[m.playerIndex]
    local k = gKingStates[m.playerIndex]

    if m.playerIndex == 0 then
        local o = obj_get_first_with_behavior_id(id_bhvKingNPC)

        if o and o.oAction == KING_NPC_ACT_IDLE
        and o.oDistanceToMario < o.hitboxRadius + 50 then
            custom_check_npc_talk(m, o)
        end

        if gSync.isKing and k.starKing.enabled == 1 then
            m.health = 0x880
            m.hurtCounter = 0
            m.invincTimer = 1
            m.peakHeight = m.pos.y
            m.healCounter = 0
        end
    end

    if (save_file_get_flags() & SAVE_FLAG_TALKED_TO_KING) ~= 0 then
        m.specialTripleJump = 0
        m.numLives = 100--max(m.numLives, 4)
    end
end

hook_event(HOOK_MARIO_UPDATE, mario_update)
hook_event(HOOK_BEFORE_MARIO_UPDATE, before_mario_update)

function dialog_sound(dialogSound)
    local m = gMarioStates[0]

    if get_dialog_id() == DIALOG_099 then
        audio_sample_play(SAMPLE_KING, m.pos, 1)
        return DS_NONE
    end
end

function check_char(char)
    if charSelectExists ~= nil then
        if charSelect.character_get_current_number() == charSelect.character_get_number_from_string(char) then
            return true
        end
    end

    if gKingStates[0].name == char then
        return true
    end

    return false
end

function on_dialog(id)

local KING_DIALOG_1 = {5, 30, 200, (
"so you really thought to\
look back here huh? that\
or you saw n-kingnpc.lua\
written clearly in the\
mod files...\
if youre confused im\
the REAL king, creator of\
the mod, and i wanted to\
give you a lil smth for\
getting all every star.\
unlike that dinosaur guy\
up there, imma let you\
CHOOSE if you want my\
cool reward or not.\
\
whats the reward you\
ask? see for yourself:\
\
\
//Accept// Decline")}

local KING_DIALOG_2 = {5, 30, 200, (
"you dont like the\
reward? i mean ye it\
might fuck up your save\
file but like come on\
its cool...\
aight well if you REALLY\
dont want it then fine,\
your choice.\
\
//Return// Keep")}

    if check_char("Sonic [SCR]") then

KING_DIALOG_1 = {3, 30, 200, (
"youre damn lucky im\
just a slightly edited\
bob-omb buddy.\
want my reward or not?\
\
//Accept// Decline")}

KING_DIALOG_2 = {4, 30, 200, (
"not this asshole again-\
gonna return the reward?\
\
//Return// Keep")}

    elseif check_char("Freddy Fazbear") then

KING_DIALOG_1 = {3, 30, 200, (
"holy shit is that\
fucking freddy fazbear??\
\
what are you doing here\
anyway? this place hella\
weird.\
jesus you found everything\
back there??? here lemme\
give you smth extra.\
you want it or not?\
\
//Accept// Decline")}

KING_DIALOG_2 = {4, 30, 200, (
"oh shit freddy uhh\
howd you like the reward?\
you fine keeping it or\
you wanna give it back?\
again your choice:\
\
\
//Return// Keep")}

    end

    if (save_file_get_flags() & SAVE_FLAG_TALKED_TO_KING) == 0 then
        smlua_text_utils_dialog_replace(DIALOG_099, 1, KING_DIALOG_1[1], KING_DIALOG_1[2], KING_DIALOG_1[3], KING_DIALOG_1[4])
    else
        smlua_text_utils_dialog_replace(DIALOG_099, 1, KING_DIALOG_2[1], KING_DIALOG_2[2], KING_DIALOG_2[3], KING_DIALOG_2[4])
    end
end

hook_event(HOOK_DIALOG_SOUND, dialog_sound)
hook_event(HOOK_ON_DIALOG, on_dialog)